#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "brs_logger.h"








#define USECS_IN_ONE_SEC 1000000
#define FALSE		0
#define TRUE		(!FALSE)



/*
 * Globals.
 */
extern int	Debug;




/*
 * Modules Globals.
 */
static char	LogDir[256];
static int	LogDirSet = FALSE;
static FILE 	*log_file = NULL;
static char	msg_buf[512];
static int		LogGeneration = 0;

static unsigned long	LogMask = 0;
static int		LogRate = 100;
static int		LogFrequency = 1;
static int		LogLookCount = 0;
static int		LogClientCount = 0;
static int		LogProcCount = 0;



static 			PurifyString (char *s, char *d);





void
LogSetMask (char *optList)
{
    LogMask = 0;

    if (strstr (optList, "all") != NULL || strlen (optList) == 0) {
	LogMask = LOG_ALL_MASK;
	return;
    }

    if (Debug >= 0) 
        LogMask |= LOG_MSG | ERROR_MSG;

    if (strstr (optList, "clie") != NULL) 
	LogMask |= CLIENT_CONNECT | CLIENT_DISCONNECT;
    if (strstr (optList, "proc") != NULL) 
	LogMask |= PROC_CREATE | PROC_EXIT;
    if (strstr (optList, "erro") != NULL) 
	LogMask |= ERROR_MSG;
    if (strstr (optList, "look") != NULL) 
	LogMask |= LOOKUP;
    if (strstr (optList, "mess") != NULL) 
	LogMask |= LOG_MSG;
}



void
LogSetRate (int rate)
{
    LogRate = rate;
    LogFrequency = 100 / rate;
    if (LogFrequency < 1) 
	LogFrequency = 1;
}



void
LogSetDir (char *ld)
{
    strcpy (LogDir, ld);
    if (LogDir[strlen(LogDir)-1] != '/')
	strcat (LogDir, "/");
    LogDirSet = TRUE;
}
    

/*
 * Called BEFORE the server forks another server process.
 * We increment the LogLookCount BECAUSE otherwise, with -n 1
 * each server process would start with LogLookCount=0 and would
 * will not log any requests until LogFrequency requests have been
 * received.
 *
 * Same with LogProcCount.  if we don't increment the value in the
 * main server process all child proccesses will start with the same
 * value, zero, and none of them will log them self (except for 100% logging).
 * 
 * But it is not done for LogClientCount because the original client record
 * is written by the main server process.
 */
void
LogPrepareToFork (void)
{
    LogLookCount = (LogLookCount + 1) % LogFrequency;
    LogProcCount = (LogProcCount + 1) % LogFrequency;
    LogFlush ();
}
    




void
LogPerror (char *msg)
{
    if ((LogMask & ERROR_MSG) == 0)
	return;

    sprintf (msg_buf, "PID=%d	%s	errno=%d", 
	     getpid(), msg, errno);
    LogMsg (ERROR_MSG, msg_buf);
}




int
LogClientConnect (unsigned int id, struct sockaddr_in *cli_addr)
{
    int		b1, b2, b3, b4;
    u_long	addr;

    if ((LogMask & CLIENT_CONNECT) == 0)
	return;

    LogClientCount = (LogClientCount + 1) % LogFrequency;
    if (LogClientCount > 0) 
	    return (FALSE);
    

    addr = htonl (cli_addr->sin_addr.s_addr);

    b4 = addr & 0xff;
    b3 = (addr & 0xff00) >> 8;
    b2 = (addr & 0xff0000) >> 16;
    b1 = (addr & 0xff000000) >> 24;

    sprintf (msg_buf, "ID=%u	host=%d.%d.%d.%d", 
	     id, b1, b2, b3, b4);

    LogMsg (CLIENT_CONNECT, msg_buf);
    return (TRUE);
}


void 
LogClientDiscon (unsigned int id, char *reason, int req_count)

{
    if ((LogMask & CLIENT_DISCONNECT) == 0)
	return;

    sprintf (msg_buf, "ID=%u	reason=%s	requests=%d", 
	     id, reason, req_count);
    LogMsg (CLIENT_DISCONNECT, msg_buf);
}



int
LogProcCreate (int pid)
{
    if ((LogMask & PROC_CREATE) == 0)
	return;

    LogProcCount = (LogProcCount + 1) % LogFrequency;
    if (LogProcCount > 0)
	    return (FALSE);

    sprintf (msg_buf, "ID=%d", pid);
    LogMsg (PROC_CREATE, msg_buf);
    return (TRUE);
}



void
LogProcExit (int pid, char *reason)
{
    if ((LogMask & PROC_EXIT) == 0)
	return;

    sprintf (msg_buf, "ID=%d	reason=%s", pid, reason);
    LogMsg (PROC_EXIT, msg_buf);
}


void
LogFOpenErr (int clientID, char *db_name, char *list_name, char *f_name, 
	     int err_errno)
{
    if ((LogMask & ERROR_MSG) == 0)
	return;

    sprintf (msg_buf, 
       "PID=%d	CID=%d	Failed to open file \"%s\"	errno=%d",
	     getpid(), clientID, f_name, err_errno);
    LogMsg (ERROR_MSG, msg_buf);
}



void
LogHeaderReadErr (char *db_name, char *list_name, char *hdr_name, 
		  int err_errno)
{
    if ((LogMask & ERROR_MSG) == 0)
	return;

    sprintf (msg_buf, 
    "PID=%d	Failed to read header from file \"%s\"	errno=%d",
	     getpid(), hdr_name, err_errno);
    LogMsg (ERROR_MSG, msg_buf);
}



void	
LogLookup (LookupStats *ls)
{
    char		db_name[256];
    char		list_name[256];
    char		prefix[128];


    if ((LogMask & LOOKUP) == 0)
	return;

    LogLookCount = (LogLookCount + 1) % LogFrequency;
    if (LogLookCount > 0) 
	return;

    PurifyString (ls->db_name, db_name);
    PurifyString (ls->list_name, list_name);

    switch (ls->lookup_type) {
    case 'N':
    case 'T':
	sprintf (msg_buf, 
"ID=%d	svct=%.3f	lookt=%.3f	db=%s	list=%s	type=%c	line=%d	lsent=%d	reads=%d",
		 ls->client_id, ((float)ls->service_time / 1000.0),
		 ((float)ls->lookup_time / 1000.0), 
		 db_name, list_name, ls->lookup_type, ls->line_req,
		 ls->lines_sent, ls->reads);
	break;

    case 'P':
	PurifyString (ls->prefix, prefix);
	sprintf (msg_buf, 
"ID=%d	svct=%.3f	lookt=%.3f	db=%s	list=%s	type=%c	prefix=%s	lsent=%d	reads=%d",
		 ls->client_id, ((float)ls->service_time / 1000.0),
		 ((float)ls->lookup_time / 1000.0), 
		 db_name, list_name, ls->lookup_type, prefix,
		 ls->lines_sent, ls->reads);
	break;

    case 'H':
	sprintf (msg_buf, 
"ID=%d	svct=%f	lookt=%f	db=%s	list=%s	type=%c	reads=%d",
		 ls->client_id, ((float)ls->service_time / 1000.0),
		 ((float)ls->lookup_time / 1000.0), 
		 db_name, list_name, ls->lookup_type, ls->reads);
	break;

    default:
	sprintf (msg_buf,
"ID=%d	svct=%f	lookt=%f	db=%s	list=%s	type=%c UNKNOWN",
		 ls->client_id, ((float)ls->service_time / 1000.0),
		 ((float)ls->lookup_time / 1000.0), 
		 db_name, list_name, ls->lookup_type);
	break;
    }
    LogMsg (LOOKUP, msg_buf);
}


void
LogOptions (void)
{
    char buf[256];
    char buf2[128];
    
    buf2[0] = '\0';
    if (LogMask & CLIENT_CONNECT)
	    strcat (buf2, " client");
    if (LogMask & PROC_CREATE)
	    strcat (buf2, " processes");
    if (LogMask & ERROR_MSG)
	    strcat (buf2, " error");
    if (LogMask & LOG_MSG)
	    strcat (buf2, " messages");
    if (LogMask & LOOKUP)
	    strcat (buf2, " lookup");
    sprintf (buf, "Logging:  %s.  Rate:  %d%%", buf2, LogRate);
    LogMsg (LOG_MSG, buf);
}


int
LogOpen (void)
{
    char buf[256];
    char buf2[128];
    struct timeval clock1;
    struct timezone tzp;
    struct tm *time1;


    if (LogMask == 0) 
	return;

    if (log_file == NULL && Debug == 0) {
	gettimeofday(&clock1, &tzp);
	time1 = localtime(&clock1.tv_sec);
	strftime(buf2, 128, "%b_%d_%H:%M:%S", time1);

	if (!LogDirSet) {
	    LogDir[0] = '\0';
	    LogDirSet = TRUE;
	}

	sprintf(buf, "%slslog_g%02d_%s_%d", LogDir, LogGeneration, 
		buf2, getpid());
	log_file = fopen(buf, "w");
	if (!log_file) {
	    return (-1);
	}
    }
    return (0);
}	




void
LogClose (void)
{
    if (Debug == 0) {
	if (log_file != NULL) {
	    fflush (log_file);
	    fclose (log_file);
	    log_file = NULL;
	}
    }
}



void
LogFlush (void)
{
    if (Debug == 0) {
	if (log_file != NULL) 
	    fflush (log_file);
    }
}




void
LogNewGeneration (void)
{
    LogClose ();
    ++LogGeneration;
}





/* Return the difference between two timevals as float in seconds */
static float
time_diff(end,start)
struct timeval end, start;
{
    struct timeval diff;

    if (end.tv_usec < start.tv_usec)
    { 
	end.tv_usec += USECS_IN_ONE_SEC; 
	end.tv_sec -= 1; 
    } 
    diff.tv_usec = end.tv_usec - start.tv_usec; 
    if (end.tv_sec < start.tv_sec)
    {
	return(0.0);
    }
    diff.tv_sec = end.tv_sec - start.tv_sec;
    return ((float) diff.tv_sec + (float) diff.tv_usec / USECS_IN_ONE_SEC);
}




/*
 * Copy string, replacing \t with <space> so that no tabs appear in the
 * output, except where they are intended as field deliminators.
 */
static 
PurifyString (char *s, char *d)
{
    char	c;

    do {
	if ((c = *d++ = *s++) == '\t')
	    *(d-1) = ' ';
    } while (c != '\0');
}


       
Tick
GetTickCount (void)
{
    struct timeval	   	t;

    gettimeofday (&t, NULL);
    return (t.tv_sec * 1000 + t.tv_usec / 1000);
}



/*
 * The following are used to convert times back and forth from 
 * floats, to integer rep of microseconds.
 */

static float mstof(ms)
int      ms;
{
    return((float)(ms / USECS_IN_ONE_SEC) + 
	   ((float)(ms % USECS_IN_ONE_SEC) / (float) USECS_IN_ONE_SEC)); 
}

static ftoms(f)
float    f;
{
    return((int) f * USECS_IN_ONE_SEC +
	   (int) ((f - (int) f) * USECS_IN_ONE_SEC));
}


void
LogMsg (int type, char *comment)
{
    char 			time_str[128];
    static struct timeval 	clock1;
    struct timezone 		tzp;
    static struct tm		*time2;
    char 			*trans_name;
    int   			mills;
    char  			mills_s[16];

    
    if (!log_file && Debug == 0) {
	if (LogOpen () != 0)
		return;
    }


    switch (type) {
    case CLIENT_CONNECT:
	trans_name = "CLNTCON";
	break;

    case CLIENT_DISCONNECT:
	trans_name = "CLNTDIS";
	break;

    case PROC_CREATE:
	trans_name = "PROCCR";
	break;

    case PROC_EXIT:
	trans_name = "PROCEX";
	break;

    case ERROR_MSG:
	trans_name = "ERROR";
	break;

    case LOG_MSG:
	if ((LogMask & LOG_MSG) == 0)
	    return;
	trans_name = "MSG";
	break;

    case LOOKUP:
	trans_name = "LOOKUP";
	break;

    default:
	trans_name = "UNKNOWN";
    }


    /* format thetime string. */
    gettimeofday(&clock1, &tzp);
    time2 = localtime(&clock1.tv_sec);
    strftime(time_str, 128, "%Y/%m/%d	%H:%M:%S", time2);
    
    /* format the milliseconds part of the string */
    mills = clock1.tv_usec / 1000;
    mills_s[0] = mills / 100 + '0';
    mills %= 100;
    mills_s[1] = mills / 10 + '0';
    mills %= 10;
    mills_s[2] = mills + '0';
    mills_s[3] = 0;


    
    if (Debug == 0)
	fprintf(log_file, "%s.%s\t%s\t%s\n", time_str, mills_s, 
	    trans_name, comment);
    else
	fprintf(stderr, "%s.%s\t%s\t\t%s\n", time_str, mills_s, 
	    trans_name, comment);
}

 
